#%%
import pathlib
from dataclasses import dataclass

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd


@dataclass
class TensileSpecimenDimensions():
    width_mm:float
    thickness_mm:float
    gauge_length_mm:float

    @property
    def csArea_mm2(self):
        """returns the cross-sectional area in mm2

        Returns:
            _type_: _description_
        """        
        return self.width_mm*self.thickness_mm



# from https://matplotlib.org/gallery/misc/cursor_demo_sgskip.html
import datetime
import pathlib
from dataclasses import dataclass
from tkinter import Tk
from tkinter.filedialog import askopenfilename

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from matplotlib.widgets import Button, TextBox
from npp_materialslab_tools.plottools import Cursor
from npp_materialslab_tools.testing_machine import TensileData, TensileSpecimenDimensions

DEVELOPMENT_FLAG = False
DEVELOPMENT_FNAME = "testingMachine/data/new XY 0 ABS_CNT 2%.csv"   


class ElasticityModulusCalcs():
    """class for performing the calculations

    Returns:
        _type_: _description_

    #TODO move this out of the file
    """    
    def __init__(self, tdobj:TensileData):
        self._testing_data_obj= tdobj
        self._df = self._testing_data_obj._data
        self.displacement = np.array(self._df .index)
        self.F_Ns = np.array(self._df ['load_avg'])

    def computations(self, specimen:TensileSpecimenDimensions, points_collected:dict):
        """perform computations

        Args:
            specimen (SpecimenDimensions): _description_
            points_collected (dict): _description_

        Returns:
            _type_: _description_
        """        
        start_ind = points_collected[1].indx
        end_ind = points_collected[2].indx
        self.csArea_mm2 = specimen.csArea_mm2
        self.exxs = self.displacement / specimen.gauge_length_mm
        E_GPa_pt = self._compute_mod_elasticity(start_ind=start_ind, end_ind=end_ind )
        ''' compute modulus of elasticity using  linear regression '''
        E_GPa_lnr = np.polyfit(self.exxs[start_ind:end_ind],self.F_Ns[start_ind:end_ind]/self.csArea_mm2,deg=1)[0]
        # print('E (pt): {:0.2f}[MPa]       E (linear regression): {:0.2f}[MPa] '.format(E_GPa_pt, E_GPa_lnr))
        return {'E_MPa_simple':E_GPa_pt, 'E_MPa_lsq':E_GPa_lnr}

    def _compute_mod_elasticity(self, start_ind, end_ind):
        '''Calculate modulus of elasticity using two points on the curve'''
        exxs  = np.zeros((2,))
        Fs  = np.zeros((2,)) # N
        
        indices = [start_ind, end_ind]
        for l in range(2):
            indx = indices[l]
            exxs[l] = self.exxs[indx]
            Fs[l] = self.F_Ns[indx] 
        E_GPa_pt = np.diff(Fs)/( self.csArea_mm2 * np.diff(exxs))
        return E_GPa_pt[0]
    
class ElasticityModulusCalculatorGUI(object):
    """
    Like Cursor but the crosshair snaps to the nearest x, y point.
    For simplicity, this assumes that *x* is sorted.

    This is a simple state machine, with the following states:
    - 0: uninitialised (ready to select first point)
    - 1: selected first point (ready to select first point)
    - 2: selected two points (perform computation)
    
    # TODO split GUI from calculations
    """

    def __init__(self,  filename=None):
        self._res_filename = None
        self._points_collected = {} # pseudo not required.

        self.fig, self.ax = plt.subplots(figsize=(12, 12))
        plt.subplots_adjust(bottom=0.25, top=0.98)
        self.txt = self.ax.text(0.7, 0.05, '', transform=self.ax.transAxes)
        
        self.fig.canvas.mpl_connect('button_press_event', self.mouse_click)

        self.load_data(filename=filename)
        self.create_plot()
        
        self.reset_SM()
        self._createTBs()

    def load_data(self, filename=None):
        """function that loads the data

        Args:
            filename (_type_, optional): _description_. Defaults to None.
        """ 
        if DEVELOPMENT_FLAG:
            filename = DEVELOPMENT_FNAME    
        if not filename:
            filename = "testingMachine/data/new XY 0 ABS_CNT 2%.csv"
            Tk().withdraw() # we don't want a full GUI, so keep the root window from appearing
            filename = askopenfilename(initialdir = "./")
        try:
            self._fname = pathlib.Path(filename)
            # self._tdobj = TestingData(fname=filename)

            self.emc = ElasticityModulusCalcs(tdobj=TensileData(fname=filename))

        except Exception as e:
            print(e)        

    @property 
    def displacement(self):
        """Delegates displacement to the ElasticityModulusCalcs

        Returns:
            _type_: _description_
        """        
        return self.emc.displacement
    @property 
    def F_Ns(self):
        """Delegates the F_Ns to the  ElasticityModulusCalcs

        Returns:
            _type_: _description_
        """        
        return self.emc.F_Ns

    def create_plot(self):
        # text location in axes coords
        self.ax.plot(self.displacement, self.F_Ns, 'o')
        self.ax.set_xlabel( 'Displacement [mm]')
        self.ax.set_ylabel( 'F[N]')

    def _createTBs(self):
        '''

        Creates textBoxes and Reset button
        plt.axes([x, y, dx, dy])
        - x from lower left corner
        - y from lower left corner
        - dx horizontal width 
        - dy Vertical width 
        '''

        row1y,row2y =   0.12, 0.02
        col1x, col2x, col3x, col4x = 0.15, 0.35, 0.65, 0.85
        btnWidth, btnHeight = 0.10, 0.05

        self.axWidthBtn = plt.axes([col1x, row1y, btnWidth, btnHeight])
        self.tbWidth = TextBox(self.axWidthBtn, 'Width [mm]', initial='10.1')
        self.axThicknessBtn = plt.axes([col2x, row1y, btnWidth, btnHeight])
        self.tbThickness = TextBox(self.axThicknessBtn, 'Thickness [mm]', initial='3.1')
        self.axGaugeLengthBtn = plt.axes([col1x, row2y, btnWidth, btnHeight])
        self.tbGaugeLength = TextBox(self.axGaugeLengthBtn, 'G. Length [mm]', initial='90')
        
        self.tbWidth.on_submit(lambda event: self.computations())
        self.tbThickness.on_submit(lambda event: self.computations())
        
        # Reset button
        self.axResetBtn = plt.axes([col4x, row1y, btnWidth, btnHeight])
        self.btnReset = Button(self.axResetBtn, 'Reset')
        self.btnReset.on_clicked(lambda event: self.reset_SM())

        # Open new data set
        self.axOpenNewFileBtn = plt.axes([col3x, row2y, btnWidth, btnHeight])
        self.btnNewFile = Button(self.axOpenNewFileBtn, 'New file')
        self.btnNewFile.on_clicked(lambda event: self.new_file())

        # append to file button
        self.axAppendToFileBtn = plt.axes([col4x, row2y, btnWidth, btnHeight])
        self.btnAppend = Button(self.axAppendToFileBtn, 'To File')
        self.btnAppend.on_clicked(lambda event: self.appendToFile())

    def new_file(self):
        """loads new file
        """        
        self.load_data(filename=None)
        self.reset_SM()

    def appendToFile(self):
        if self._sm != 2:
            print ('Cannot estimate E yet!')
            return 
        if not self._res_filename:
            self._res_filename = datetime.datetime.now().strftime("res%Y%m%d%H%M%S.txt")
            #self._res_filename = 'results_log.txt'
            with open(self._res_filename, 'w') as file:
                file.write('{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}\n'.format('Timestamp','Filename','Width', 'Thickness', 'start_Ind', 'end_Ind', 'E_GPA_pt', 'E_GPA_lrg'))
       
        with open(self._res_filename, 'a') as file:
            E_GPa_pt = self._compute_mod_elasticity()
            E_GPa_lnr = self._compute_mod_elasticity_lnr(self._points_collected[1].indx, self._points_collected[2].indx)
            file.write('{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}\n'.format(datetime.datetime.now().strftime("%Y%m%d-%H%M%S"), self._fname.stem ,self._dim_w, self._dim_t, self._points_collected[1].indx, self._points_collected[2].indx, E_GPa_pt, E_GPa_lnr))
       
    def reset_SM(self):
        ''' resets state machine status'''
        try:
            for i in self._points_collected.keys():
                self._points_collected[i].remove_line()
            self.ax.lines.remove(self._l_selected[0])
            self._l_selected = []
        except Exception as e:
            print (e)
        finally:
            self.ax.figure.canvas.draw() 

        self._points_collected = {}
        self._sm = 0
        self.txt.set_text('')

        self.ax.cla()
        self.create_plot()
        self.txt = self.ax.text(0.01, 0.95, self._fname.stem, transform=self.ax.transAxes)

    def get_specimenDimensions(self):
        ''' calcate crossection'''
        self._dim_w = float(self.tbWidth.text)
        self._dim_t = float(self.tbThickness.text)
        self._dim_l = float(self.tbGaugeLength.text)

        self._specimenDimensions = TensileSpecimenDimensions(gauge_length_mm=self._dim_l, thickness_mm= self._dim_t, width_mm=self._dim_w)
        return self._specimenDimensions

    def computations(self):
        specimenDimensions= self.get_specimenDimensions()
        self.exxs = self.displacement / specimenDimensions.gauge_length_mm
        res= self.emc.computations(specimen=specimenDimensions, points_collected = self._points_collected)
        E_GPa_pt = res['E_MPa_simple']
        E_GPa_lnr = res['E_MPa_lsq']
        #output to plot and console
        start_ind = self._points_collected[1].indx
        end_ind = self._points_collected[2].indx
        self._l_selected = self.ax.plot(self.displacement[start_ind:end_ind], self.F_Ns[start_ind:end_ind], 'r')
        self.txt = self.ax.text(0.7, 0.05, '', transform=self.ax.transAxes)
        self.txt.set_text('E = {:.2f}[Mpa]'.format (E_GPa_lnr))
        # console
        print('E (pt): {:0.2f}[MPa]       E (linear regression): {:0.2f}[MPa] '.format(E_GPa_pt, E_GPa_lnr))

    def _calc_x_y_ind(self, x_event, y_event ):
        ''' calculates the x, y and index from the event data

        It does that by finding the closest point.

        Callers: mouse_click
        '''
        indx = min(np.searchsorted(self.displacement, x_event), len(self.displacement) - 1)
        x = self.displacement[indx]
        y = self.F_Ns[indx]
        return x, y, indx       

    def mouse_click(self, event):
        ''' change the state machine on each click
        ''' 
        if not event.inaxes:
            ''' only continue when a point is picked'''
            return

        # update state machine
        if self._sm == 0:
            crsrID  = self._sm+1
            x,y, indx = self._calc_x_y_ind(event.xdata, event.ydata)
            # update the line positions
            self._points_collected[crsrID] = Cursor(self.ax, crsrID, x, y, indx)
            self._points_collected[crsrID].plot_cursor()
            self._sm = 1
        elif self._sm == 1:
            crsrID  = self._sm+1
            x,y, indx = self._calc_x_y_ind(event.xdata, event.ydata)
            # update the line positions
            self._points_collected[crsrID] = Cursor(self.ax, crsrID, x, y, indx)
            self._points_collected[crsrID].plot_cursor()
            
            self._sm = 2
        if self._sm ==2:
            self.computations()
        else:
            print ("State status: {} | x={:1.2f}, y={:1.2f}".format(self._sm,x,y))
        self.ax.figure.canvas.draw()



#%%
if __name__ == "__main__":

    # snap_cursor = ElasticityModulusCalculator('../Results/20190327-tests/D00_02.xlsx')
    snap_cursor = ElasticityModulusCalculatorGUI()

    plt.show()


#%%[markdown]
# this is a copy of the pydic file. 
# 
# For installation requirements check website


#%%[markdown]
#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This file is part of pydic, a free digital correlation suite for computing strain fields
#
# Author :  - Damien ANDRE, SPCTS/ENSIL-ENSCI, Limoges France
#             <damien.andre@unilim.fr>
#
# Copyright (C) 2017 Damien ANDRE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Contributors : Ronan GARO (debugging + log strain), Laurent MAHEO (debugging + log strain)

# ====== INTRODUCTION
# Welcome to pydic a free python suite for digital image correlation.
# pydic allows to compute (smoothed or not) strain fields from a serie of pictures.
# pydic takes in account the rigid body transformation.

# Note that this file is a module file, you can't execute it.
# You can go to the example directory for usage examples.


import copy
import glob
import logging
import math
import os
import sys

import cv2
import numpy as np
from matplotlib import pyplot as plt
from matplotlib.widgets import Button, RadioButtons, Slider

from .pydicGrid import grid, draw_opencv


# logging.basicConfig(level=logging.DEBUG)


      
def build_grid(area, num_point, *args, **kwargs):
     """
     TODO: Findout what are the callers for this function from the original repository

     Args:
         area (_type_): _description_
         num_point (_type_): _description_

     Returns:
         _type_: _description_
     """    
     xmin = area[0][0]; xmax = area[1][0]; dx = xmax - xmin
     ymin = area[0][1]; ymax = area[1][1]; dy = ymax - ymin
     point_surface = dx*dy/num_point; point_line = math.sqrt(point_surface)
     ratio = 1. if not 'ratio' in kwargs else kwargs['ratio']
     num_x = int(ratio*dx/point_line) + 1
     num_y = int(ratio*dy/point_line) + 1
     grid_x, grid_y = np.mgrid[xmin:xmax:num_x*1j, ymin:ymax:num_y*1j]
     return grid(grid_x, grid_y, num_x, num_y)

def write_result(result_file, image, points):
     """used by init to write the data for a file.

     Args:
         result_file (_type_): _description_
         image (_type_): _description_
         points (_type_): _description_
     """     
     result_file.write(image + '\t')
     for p in points:
          result_file.write(str(p[0]) + ',' + str(p[1]) + '\t')
     result_file.write('\n')
    
def init(image_pattern, win_size_px, grid_size_px, result_file, area_of_interest=None, *args, **kwargs):
     """the init function is a simple wrapper function that allows to parse a 
sequence of images. The displacements are computed and a result file is written

     Args:
         image_pattern (str): the path and pattern describing where your image are located 
         win_size_px (list):  the size in pixel of your correlation windows.Given as a (dx, dy) tuple
         grid_size_px (list): the size of your correlation grid. Given as a (dx, dy) tuple
         result_file (str): the name of the result file
         area_of_interest (list of two tuples, optional): gives the area of interset in 
               [(top left x,top left xy),(bottom right x, bottom right y)  ] format.
                 Defaults to None.
                         if you don't give this argument, a windows with the first image is displayed. 
                         You can pick in this picture manually your area of interest.

     Parsed kwargs:
          unstructured_grid=(val1,val2) : to let the 'goodFeaturesToTrack' opencv2 algorithm. Note that you can't use the 'spline' or the 'raw' interpolation method.
     """

     
     img_list = sorted(glob.glob(image_pattern))
     assert len(img_list) > 1, "there is not image in " + str(image_pattern)
     img_ref = cv2.imread(img_list[0], 0)
     
     # choose area of interset 
     if (area_of_interest is None):
          print("please pick your area of interest on the picture")
          print("Press 'c' to proceed")
          area_of_interest = pick_area_of_interest(img_ref)

     # init correlation grid
     area     = area_of_interest

     points   = []
     points_x = np.float64(np.arange(area[0][0], area[1][0], grid_size_px[0]))
     points_y = np.float64(np.arange(area[0][1], area[1][1], grid_size_px[1]))

     if 'unstructured_grid' in kwargs:
          block_size, min_dist = kwargs['unstructured_grid']
          feature_params = dict( maxCorners = 50000,
                                 qualityLevel = 0.01,
                                 minDistance = min_dist,
                                 blockSize = block_size)
          points = cv2.goodFeaturesToTrack(img_ref, mask = None, **feature_params)[:,0]
     elif 'deep_flow' in kwargs:
          points_x = np.float64(np.arange(area[0][0], area[1][0], 1))
          points_y = np.float64(np.arange(area[0][1], area[1][1], 1))
          for x in points_x:
               for y in points_y:
                    points.append(np.array([np.float32(x),np.float32(y)]))
          points = np.array(points)
     else: 
          for x in points_x:
               for y in points_y:
                    points.append(np.array([np.float32(x),np.float32(y)]))
          points = np.array(points)


     # ok, display
     points_in = remove_point_outside(points, area, shape='box')

     
     img_ref = cv2.imread(img_list[0], 0)
     img_ref = cv2.putText(img_ref, "Displaying markers... Press any buttons to continue", (50,50), cv2.FONT_HERSHEY_SIMPLEX, 1,(255,255,255),4)

     draw_opencv(img_ref, point=points_in)

     # compute grid and save it in result file
     f = open(result_file, 'w')
     xmin = points_x[0]; xmax = points_x[-1]; xnum = len(points_x)
     ymin = points_y[0]; ymax = points_y[-1]; ynum = len(points_y)
     f.write(str(xmin) + '\t' + str(xmax) + '\t' + str(int(xnum)) + '\t' + str(int(win_size_px[0])) + '\n')
     f.write(str(ymin) + '\t' + str(ymax) + '\t' + str(int(ynum)) + '\t' + str(int(win_size_px[1])) + '\n')

     # param for correlation 
     lk_params = dict( winSize  = win_size_px, maxLevel = 10,
                       criteria = (cv2.TERM_CRITERIA_EPS | cv2.TERM_CRITERIA_COUNT, 10, 0.03))
     
     # parse all files and write results file
     point_to_process = points_in
     write_result(f, img_list[0], point_to_process)
     for i in range(len(img_list)-1):
          print('reading image {} / {} : "{}"'.format(i+1, len(img_list), img_list[i+1]))
          image_ref = cv2.imread(img_list[i], 0)
          image_str = cv2.imread(img_list[i+1], 0)
          
          if 'deep_flow' in kwargs:
               winsize_x = win_size_px[0]
               final_point = cv2.calcOpticalFlowFarneback(image_ref, image_str, None, 0.5, 3, winsize_x,
                                                          10, 5, 1.2, 0)
               # prev, next, flow, pyr_scale, levels, winsize, iterations,poly_n, poly_sigma
               index = 0
               ii_max = final_point.shape[0]
               jj_max = final_point.shape[1]

               for jj in range(jj_max):
                   for ii in range(ii_max):
                      #area     = [(0,0),(img_ref.shape[1],img_ref.shape[0])]
                      if (jj >= area[0][0] and jj < area[1][0] and
                          ii >= area[0][1] and ii < area[1][1]):
                          point_to_process[index] += final_point[ii,jj]
                          index += 1

               
          else:
               final_point, st, err = cv2.calcOpticalFlowPyrLK(image_ref, image_str, point_to_process, None, **lk_params)               
               #draw_opencv(image_ref, point=points_in, pointf=final_point, l_color=(0,255,0), p_color=(0,255,0))
               point_to_process = final_point
          write_result(f, img_list[i+1], point_to_process)
     f.write('\n')
     f.close()


def read_dic_file(result_file, *args, **kwargs):
     """the read_dic_file is a simple wrapper function that allows to parse a dic 
file (given by the init() function) and compute the strain fields. The displacement fields 
can be smoothed thanks to many interpolation methods. A good interpolation method to do this 
job is the 'spline' method. After this process, note that a new folder named 'pydic' is 
created into the image directory where different results files are written. 

These results are :
- 'disp' that contains images where the displacement of the correlation windows are highlighted. You 
  can apply a scale to amplify these displacements.
- 'grid' that contains images where the correlation grid is highlighted. You 
  can apply a scale to amplify the strain of this grid.
- 'marker' that contains images  where the displacement of corraleted markers are highlighted
- 'result' where you can find raw text file (csv format) that constain the computed displacement
  and strain fields of each picture.

* required argument:
  - the first arg 'result_file' must be a result file given by the init() function
* optional named arguments ;
 - 'interpolation' the allowed vals are 'raw', 'spline', 'linear', 'delaunnay', 'cubic', etc... 
   a good value is 'raw' (for no interpolation) or spline that smooth your data.
 - 'save_image ' is True or False. Here you can choose if you want to save the 'disp', 'grid' and 
   'marker' result images
 - 'scale_disp' is the scale (a float) that allows to amplify the displacement of the 'disp' images
 - 'scale_grid' is the scale (a float) that allows to amplify the 'grid' images
 - 'meta_info_file' is the path to a meta info file. A meta info file is a simple csv file 
   that contains some additional data for each pictures such as time or load values.
 - 'strain_type' should be 'cauchy' '2nd_order' or 'log'. Default value is cauchy (or engineering) strains. You 
   can switch to log or 2nd order strain if you expect high strains. 
 - 'rm_rigid_body_transform' for removing rigid body displacement (default is true)
"""
     grid_list = [] # saving grid here
     # treat optional args
     interpolation= 'raw' if not 'interpolation' in kwargs else kwargs['interpolation']
     save_image   = True if not 'save_image' in kwargs else kwargs['save_image']
     scale_disp   = 4. if not 'scale_disp' in kwargs else float(kwargs['scale_disp'])
     scale_grid   = 25. if not 'scale_grid' in kwargs else float(kwargs['scale_grid'])
     strain_type  = 'cauchy' if not 'strain_type' in kwargs else kwargs['strain_type']
     rm_rigid_body_transform = True if not 'rm_rigid_body_transform' in kwargs else kwargs['rm_rigid_body_transform']
     
     # read meta info file
     meta_info = {}
     if 'meta_info_file' in kwargs:
          print('read meta info file', kwargs['meta_info_file'], '...')
          with open(kwargs['meta_info_file']) as f:
               lines = f.readlines()
               header = lines[0]
               field = header.split()
               for l in lines[1:-1]:
                    val = l.split()
                    if len(val) > 1:
                         dictionary = dict(zip(field, val))
                         meta_info[val[0]] = dictionary
     
                
     # first read grid
     with open(result_file) as f:
          head = f.readlines()[0:2]
     (xmin, xmax, xnum, win_size_x) = [float(x) for x in head[0].split()]
     (ymin, ymax, ynum, win_size_y) = [float(x) for x in head[1].split()]
     win_size = (win_size_x, win_size_y)
     
     grid_x, grid_y = np.mgrid[xmin:xmax:int(xnum)*1j, ymin:ymax:int(ynum)*1j]
     mygrid = grid(grid_x, grid_y, int(xnum), int(ynum))

     # the results
     point_list = []
     image_list = []
     disp_list = []

     # parse the result file
     with open(result_file) as f:
          res = f.readlines()[2:-1]
          for line in res:
               val = line.split('\t')
               image_list.append(val[0])
               point = []
               for pair in val[1:-1]:
                    (x,y) = [float(x) for x in pair.split(',')]
                    point.append(np.array([np.float32(x),np.float32(y)]))
               point_list.append(np.array(point))
               grid_list.append(copy.deepcopy(mygrid))
     f.close()
               
     # compute displacement and strain
     for i, mygrid in enumerate(grid_list):
          print("compute displacement and strain field of", image_list[i], "...")
          disp = None
          if rm_rigid_body_transform:
               print("remove rigid body transform")
               disp = compute_disp_and_remove_rigid_transform(point_list[i], point_list[0])
          else:
               print("do not remove rigid body transform")
               disp = compute_displacement(point_list[i], point_list[0])
          mygrid.add_raw_data(win_size, image_list[0], image_list[i], point_list[0], point_list[i], disp)
          
          disp_list.append(disp)
          mygrid.interpolate_displacement(point_list[0], disp, method=interpolation)

          if (strain_type == 'cauchy'):
               mygrid.compute_strain_field()
          elif (strain_type =='2nd_order'):
               mygrid.compute_strain_field_DA()
          elif (strain_type =='log'):
               mygrid.compute_strain_field_log()
          else:
               print("please specify a correct strain_type : 'cauchy', '2nd_order' or 'log'")
               print("exiting...")
               sys.exit(0)

          # write image files
          if (save_image):
               mygrid.draw_marker_img()
               mygrid.draw_disp_img(scale_disp)
               mygrid.draw_grid_img(scale_grid)
               if win_size_x == 1 and win_size_y == 1 : 
                    mygrid.draw_disp_hsv_img()

          # write result file
          mygrid.write_result()

          # add meta info to grid if it exists
          if (len(meta_info) > 0):
               img = os.path.basename(mygrid.image)
               #if not meta_info.has_key(img):
               if img not in meta_info.keys():
                    print("warning, can't affect meta deta for image", img)
               else:
                    mygrid.add_meta_info(meta_info.get(img))
                    print('add meta info', meta_info.get(img))
     return grid_list
                    
          

def compute_displacement(point, pointf):
    """To compute a displacement between two point arrays"""
    assert len(point)==len(pointf)
    values = []
    for i, pt0 in enumerate(point):
        pt1 = pointf[i]
        values.append((pt1[0]-pt0[0], pt1[1]-pt0[1]))
    return values

# TODO: Move those two into the pick_area_of_interest
# this should remove the need for 
area = []
cropping = False # I don't see much use in the cropping flag. 

def pick_area_of_interest(PreliminaryImage):
     """used by init. Picks area of interest from an image
     if data are not selected


     Args:
         PreliminaryImage (_type_): _description_

     Returns:
         _type_: _description_
     """    
     global area, cropping
     image = cv2.putText(PreliminaryImage, "Pick the area of interest (left click + move mouse) and press 'c' button to continue", (50,50), cv2.FONT_HERSHEY_SIMPLEX, 1,(255,255,255),4)
          
     def click_and_crop(event, x, y, flags, param):
          global area, cropping
          if event == cv2.EVENT_LBUTTONDOWN:
               area = [(x, y)]
               cropping = True

          elif event == cv2.EVENT_LBUTTONUP:
               area.append((x, y))
               cropping = False

               # draw a rectangle around the region of interest
               Newimage = cv2.rectangle(image, area[0], area[1], (0, 255, 0), 2)
               cv2.imshow('image', Newimage)
               
     clone = image.copy()
     cv2.namedWindow('image', cv2.WINDOW_NORMAL)
     cv2.resizeWindow('image', image.shape[1], image.shape[0])
     cv2.setMouseCallback("image", click_and_crop)

     # keep looping until the 'c' key is pressed
     while True:
          # display the image and wait for a keypress
          cv2.imshow("image", image)
          key = cv2.waitKey(1) & 0xFF

          # if the 'r' key is pressed, reset the cropping region
          if key == ord("r"):
               image = clone.copy()

               # if the 'c' key is pressed, break from the loop
          elif key == ord("c"):
               print(f'The area of interest is {area}')
               break
     return area

def remove_point_outside(points, area,  *args, **kwargs):
     shape = 'box' if not 'shape' in kwargs else kwargs['shape']
     # what is shape doing here?
     xmin = area[0][0]; xmax = area[1][0]
     ymin = area[0][1]; ymax = area[1][1]
     res = []
     for p in points:
          x = p[0]; y = p[1]
          if ((x >= xmin) and (x <= xmax) and (y >= ymin) and (y <= ymax)):
               res.append(p)
     return np.array(res)


def compute_disp_and_remove_rigid_transform(p1, p2):
     """_summary_

     Args:
         p1 (_type_): _description_
         p2 (_type_): _description_

     Returns:
         _type_: _description_
     """     
     A = []
     B = []
     removed_indices = []
     for i in range(len(p1)):
          if np.isnan(p1[i][0]):
               assert np.isnan(p1[i][0]) and np.isnan(p1[i][1]) and np.isnan(p2[i][0]) and np.isnan(p2[i][1])
               removed_indices.append(i)
          else:
               A.append(p1[i])
               B.append(p2[i])
          
     A = np.matrix(A)
     B =  np.matrix(B)
     assert len(A) == len(B)
     N = A.shape[0]; # total points
     
     centroid_A = np.mean(A, axis=0)
     centroid_B = np.mean(B, axis=0)
    
     # centre the points
     AA = np.matrix(A - np.tile(centroid_A, (N, 1)))
     BB = np.matrix(B - np.tile(centroid_B, (N, 1)))

     # dot is matrix multiplication for array
     H = np.transpose(AA) * BB
     U, S, Vt = np.linalg.svd(H)
     R = Vt.T * U.T

     # special reflection case
     if np.linalg.det(R) < 0:
          print("Reflection detected")
          Vt[2,:] *= -1
          R = Vt.T * U.T

     n = len(A)
     T = -R*centroid_A.T + centroid_B.T
     A2 = (R*A.T) + np.tile(T, (1, n))
     A2 = np.array(A2.T)
     out = []
     j = 0
     for i in range(len(p1)):
          if np.isnan(p1[i][0]):
               out.append(p1[i])
          else:
               out.append(A2[j])
               j = j + 1
     out = np.array(out)
     return compute_displacement(p2, out)





# from .appElasticityCalculator import ElasticityModulusCalcs
from npp_materialslab_tools.plottools.cursor_crosshairs import Cursor
from npp_materialslab_tools.plottools.point_selector import PointSelector
from npp_materialslab_tools.plottools.point_selector2 import PointsSelector2


import math
import os

import cv2
import numpy as np
import scipy.interpolate
from matplotlib import pyplot as plt
from matplotlib.widgets import Button, RadioButtons, Slider
from scipy.interpolate import Rbf, griddata


class Plot:
    def __init__(self, image, grid, data, title):
        self.data = np.ma.masked_invalid(data)
        self.data_copy = np.copy(self.data)
        self.grid_x = grid.grid_x
        self.grid_y = grid.grid_y
        self.data = np.ma.array(self.data, mask=self.data == np.nan)
        self.title = title
        self.image = image

        self.fig = plt.figure()
        self.ax = self.fig.add_subplot(111)
        self.fig.subplots_adjust(left=0.25, bottom=0.25)

        self.ax.imshow(image, cmap=plt.cm.binary)
        #ax.contour(grid_x, grid_y, g, 10, linewidths=0.5, colors='k', alpha=0.7)

        
        self.im = self.ax.contourf(grid.grid_x, grid.grid_y, self.data, 10, cmap=plt.cm.rainbow,
                         vmax=self.data.max(), vmin=self.data.min(), alpha=0.7)
        self.contour_axis = plt.gca()

        self.ax.set_title(title)
        self.cb = self.fig.colorbar(self.im)

        axmin = self.fig.add_axes([0.25, 0.1, 0.65, 0.03])
        axmax = self.fig.add_axes([0.25, 0.15, 0.65, 0.03])
        self.smin = Slider(axmin, 'set min value', self.data.min(), self.data.max(), valinit=self.data.min(),valfmt='%1.6f')
        self.smax = Slider(axmax, 'set max value', self.data.min(), self.data.max(), valinit=self.data.max(),valfmt='%1.6f')
        
        self.smax.on_changed(self.update)
        self.smin.on_changed(self.update)
        

    def update(self, val):
        self.contour_axis.clear()
        self.ax.imshow(self.image, cmap=plt.cm.binary)
        self.data = np.copy(self.data_copy)
        self.data = np.ma.masked_where(self.data > self.smax.val, self.data)
        self.data = np.ma.masked_where(self.data < self.smin.val, self.data)
        self.data = np.ma.masked_invalid(self.data)

        self.im = self.contour_axis.contourf(self.grid_x, self.grid_y, self.data, 10, cmap=plt.cm.rainbow, alpha=0.7)

        
        self.cb.update_bruteforce(self.im)
        self.cb.set_clim(self.smin.val, self.smax.val)
        self.cb.set_ticks(np.linspace(self.smin.val, self.smax.val, num=10))


        # # self.cb = self.figure.colorbar(self.im)


        # self.cb.set_clim(self.smin.val, self.smax.val)
        # self.cb.on_mappable_changed(self.im)
        # self.cb.draw_all() 
        # self.cb.update_normal(self.im)
        # self.cb.update_bruteforce(self.im)
        #plt.colorbar(self.im)


def draw_opencv(image, *args, **kwargs):
    """A function with a lot of named argument to draw opencv image
     - 'point' arg must be an array of (x,y) point
     - 'p_color' arg to choose the color of point in (r,g,b) format
     - 'pointf' to draw lines between point and pointf, pointf 
     must be an array of same lenght than the point array
     - 'l_color' to choose the color of lines
     - 'grid' to display a grid, the grid must be a grid object
     - 'gr_color' to choose the grid color"""
    if type(image) == str :
         image = cv2.imread(image, 0)

    if 'text' in kwargs:
         text = kwargs['text']
         image = cv2.putText(image, text, (50,50), cv2.FONT_HERSHEY_SIMPLEX, 1,(255,255,255),4)

         
    frame = cv2.cvtColor(image, cv2.COLOR_GRAY2BGR)
    if  'point' in kwargs:
        p_color = (0, 255, 255) if not 'p_color' in kwargs else kwargs['p_color']
        for pt in kwargs['point']:
            if not np.isnan(pt[0]) and not np.isnan(pt[1]):
                 x = int(pt[0])
                 y = int(pt[1])
                 frame = cv2.circle(frame, (x, y), 4, p_color, -1)

    scale = 1. if not 'scale' in kwargs else kwargs['scale']
    if 'pointf' in kwargs and 'point' in kwargs:
        assert len(kwargs['point']) == len(kwargs['pointf']), 'bad size'
        l_color = (255, 120, 255) if not 'l_color' in kwargs else kwargs['l_color']
        for i, pt0 in enumerate(kwargs['point']):
            pt1 = kwargs['pointf'][i]
            if np.isnan(pt0[0])==False and np.isnan(pt0[1])==False and np.isnan(pt1[0])==False and np.isnan(pt1[1])==False :
                 disp_x = (pt1[0]-pt0[0])*scale
                 disp_y = (pt1[1]-pt0[1])*scale
                 frame = cv2.line(frame, (int(pt0[0]), int(pt0[1])), (int(pt0[0]+disp_x), int(pt0[1]+disp_y)), l_color, 2)

    if 'grid' in kwargs:
        gr =  kwargs['grid']
        gr_color = (255, 255, 255) if not 'gr_color' in kwargs else kwargs['gr_color']
        for i in range(gr.size_x):
            for j in range(gr.size_y):
                 if (not math.isnan(gr.grid_x[i,j]) and  
                     not math.isnan(gr.grid_y[i,j]) and
                     not math.isnan(gr.disp_x[i,j]) and  
                     not math.isnan(gr.disp_y[i,j])):
                      x = int(gr.grid_x[i,j]) - int(gr.disp_x[i,j]*scale)
                      y = int(gr.grid_y[i,j]) - int(gr.disp_y[i,j]*scale)
                      
                      if i < (gr.size_x-1):
                           if (not math.isnan(gr.grid_x[i+1,j]) and  
                               not math.isnan(gr.grid_y[i+1,j]) and
                               not math.isnan(gr.disp_x[i+1,j]) and  
                               not math.isnan(gr.disp_y[i+1,j])):
                                x1 = int(gr.grid_x[i+1,j]) - int(gr.disp_x[i+1,j]*scale)
                                y1 = int(gr.grid_y[i+1,j]) - int(gr.disp_y[i+1,j]*scale)
                                frame = cv2.line(frame, (x, y), (x1, y1), gr_color, 2)

                      if j < (gr.size_y-1):
                           if (not math.isnan(gr.grid_x[i,j+1]) and  
                               not math.isnan(gr.grid_y[i,j+1]) and
                               not math.isnan(gr.disp_x[i,j+1]) and  
                               not math.isnan(gr.disp_y[i,j+1])):
                                x1 = int(gr.grid_x[i,j+1]) - int(gr.disp_x[i,j+1]*scale)
                                y1 = int(gr.grid_y[i,j+1]) - int(gr.disp_y[i,j+1]*scale)
                                frame = cv2.line(frame, (x, y), (x1, y1), gr_color, 4)
    if 'filename' in kwargs:
         cv2.imwrite( kwargs['filename'], frame)
         return

    cv2.namedWindow('image', cv2.WINDOW_NORMAL)
    cv2.resizeWindow('image', frame.shape[1], frame.shape[0])
    cv2.imshow('image',frame)
    cv2.waitKey(0)
    cv2.destroyAllWindows()
  
class grid:
     """The grid class is the main class of pydic. This class embed a lot of usefull
method to treat and post-treat results"""
     
     def __init__(self, grid_x, grid_y, size_x, size_y):
          """Construct a new grid object with 

          Args:
              grid_x (np.ndarray): x coordinate (grid_x) for each marker
              grid_y (np.ndarray): y coordinate (grid_x) for each marker
              size_x (int): number of point along x (size_x)
              size_y (int): number of point along y (size_y)
          """          
          self.grid_x = grid_x
          self.grid_y = grid_y
          self.size_x = size_x
          self.size_y = size_y
          self.disp_x =  self.grid_x.copy().fill(0.)
          self.disp_y =  self.grid_y.copy().fill(0.)
          self.strain_xx = None
          self.strain_yy = None
          self.strain_xy = None

     def add_raw_data(self, winsize, reference_image, image, reference_point, correlated_point, disp):
          """Save raw data to the current grid object. 
          
          These raw data are used as initial data 
          for digital image correlation

          Args:
              winsize (tupe): the size in pixel of the correlation windows
              reference_image (str): filename of the reference image
              image (sr): filename of the current image
              reference_point (np.ndarray[(size_x*size_y),2]): Reference coordinates for each marker 
              correlated_point (np.ndarray[(size_x*size_y),2]): Current coordinate for each marker
              disp (list of tuples): ???? 
          """          

          self.winsize = winsize
          self.reference_image = reference_image
          self.image = image
          self.reference_point = reference_point
          self.correlated_point = correlated_point
          self.disp = disp

     def add_meta_info(self, meta_info):
          """Save the related meta info into the current grid object"""
          self.meta_info = meta_info

     def prepare_saved_file(self, prefix, extension):
          """prepares the filename in the form:

          <image folder>/pydic/<prefix>/<image_name>_<prefix>.<extension>

          Args:
              prefix (str): forder that the file will be saved in the pydic folder structure
              extension (str): File extension

          Returns:
              _type_: _description_
          """

          folder = os.path.dirname(self.image)
          folder = folder + '/pydic/' + prefix
          if not os.path.exists(folder):os.makedirs(folder)
          base = os.path.basename(self.image)
          name = folder + '/' + os.path.splitext(base)[0] + '_' + prefix + '.' + extension
          print("saving", name, "file...")
          return name

     def draw_marker_img(self):
          """Draw marker image"""
          name = self.prepare_saved_file('marker', 'png')
          draw_opencv(self.image, point=self.correlated_point, l_color=(0,0,255), p_color=(255,255,0), filename=name, text=name)
          
     def draw_disp_img(self, scale):
          """Draw displacement image. A scale value can be passed to amplify the displacement field"""
          name = self.prepare_saved_file('disp', 'png')
          draw_opencv(self.reference_image, point=self.reference_point, pointf=self.correlated_point, l_color=(0,0,255), p_color=(255,255,0), scale=scale, filename=name, text=name)


                    
     def draw_disp_hsv_img(self, *args, **kwargs):
          """Draw displacement image in a hsv view."""
          name = self.prepare_saved_file('disp_hsv', 'png')
          img = self.reference_image
          if type(img) == str :
               img = cv2.imread(img, 0)
          

          disp = self.correlated_point - self.reference_point          
          fx, fy = disp[:,0], disp[:,1]
          v_all = np.sqrt(fx*fx + fy*fy)
          v_max = np.mean(v_all) + 2.*np.std(v_all)

          
          rgb = cv2.cvtColor(img, cv2.COLOR_GRAY2BGR)
          hsv = cv2.cvtColor(rgb, cv2.COLOR_BGR2HSV)

          if v_max != 0.:
               for i, val in enumerate(self.reference_point):
                    disp = self.correlated_point[i] - val
                    ang = np.arctan2(disp[1], disp[0]) + np.pi
                    v = np.sqrt(disp[0]**2 + disp[1]**2)
                    pt_x = int(val[0])
                    pt_y = int(val[1])

                    hsv[pt_y,pt_x, 0] = int(ang*(180/np.pi/2))
                    hsv[pt_y,pt_x, 1] = 255 if int((v/v_max)*255.) > 255 else int((v/v_max)*255.)
                    hsv[pt_y,pt_x, 2] = 255 if int((v/v_max)*255.) > 255 else int((v/v_max)*255.)


          bgr = cv2.cvtColor(hsv, cv2.COLOR_HSV2BGR)
          bgr = cv2.putText(bgr, name, (50,50), cv2.FONT_HERSHEY_SIMPLEX, 1,(255,255,255),4)

          if 'save_img' in kwargs:
               cv2.imwrite(name, bgr)
          if 'show_img' in kwargs:
               cv2.namedWindow('image', cv2.WINDOW_NORMAL)
               cv2.resizeWindow('image', bgr.shape[1], bgr.shape[0])
               cv2.imshow('image', bgr)
               cv2.waitKey(0)
               cv2.destroyAllWindows()




     def draw_grid_img(self, scale):
          """Draw grid image. A scale value can be passed to amplify the displacement field"""
          name = self.prepare_saved_file('grid', 'png')
          draw_opencv(self.reference_image, grid = self, scale=scale, gr_color=(255,255,250), filename=name, text=name)

     def write_result(self):
          """write a raw csv result file. Indeed, you can use your favorite tool to post-treat this file"""
          name = self.prepare_saved_file('result', 'csv')
          f = open(name, 'w')
          f.write("index" + ',' + "index_x" + ',' + "index_y" + ',' + "pos_x"    + ',' + "pos_y"    + ',' + 
                  "disp_x"    + ',' + "disp_y"    + ',' + 
                  "strain_xx" + ',' + "strain_yy" + ',' + "strain_xy" + '\n')
          index = 0
          for i in range(self.size_x):
            for j in range(self.size_y):
                 f.write(str(index)                                                 + ',' +
                         str(i)                   + ',' + str(j)                   + ',' + 
                         str(self.grid_x[i,j])    + ',' + str(self.grid_y[i,j])    + ',' + 
                         str(self.disp_x[i,j])    + ',' + str(self.disp_y[i,j])    + ',' + 
                         str(self.strain_xx[i,j]) + ',' + str(self.strain_yy[i,j]) + ',' + str(self.strain_xy[i,j]) + '\n')
                 index = index + 1
          f.close()


     def plot_field(self, field, title):
          """Plot the chosen field 
          such as strain_xx, disp_xx, etc. 
          in a matplotlib interactive map"""
          image_ref = cv2.imread(self.image, 0)
          Plot(image_ref, self, field, title)
          
     def interpolate_displacement(self, point, disp, *args, **kwargs):
          """Interpolate the displacement field. It allows to (i) construct the displacement grid and to 
(ii) smooth the displacement field thanks to the chosen method (raw, linear, spline,etc.)"""

          x = np.array([p[0] for p in point])
          y = np.array([p[1] for p in point])
          dx = np.array([d[0] for d in disp])
          dy = np.array([d[1] for d in disp])
          method = 'linear' if not 'method' in kwargs else kwargs['method']

          print('interpolate displacement with', method, 'method')
          if method=='delaunay':
               from scipy.interpolate import LinearNDInterpolator
               inter_x = LinearNDInterpolator(point, dx)
               inter_y = LinearNDInterpolator(point, dy)
               self.disp_x = inter_x(self.grid_x, self.grid_y)
               self.disp_y = inter_y(self.grid_x, self.grid_y)
               
          elif method=='raw':
               # need debugging
               self.disp_x = self.grid_x.copy()
               self.disp_y = self.grid_y.copy()

               assert self.disp_x.shape[0] == self.disp_y.shape[0], "bad shape"
               assert self.disp_x.shape[1] == self.disp_y.shape[1], "bad shape"
               assert len(dx) == len(dy), "bad shape"
               assert self.disp_x.shape[1]*self.disp_x.shape[0] == len(dx), "bad shape"
               count = 0
               for i in range(self.disp_x.shape[0]):
                    for j in range(self.disp_x.shape[1]):
                         self.disp_x[i,j] = dx[count]
                         self.disp_y[i,j] = dy[count]
                         count = count + 1
                         
          elif method=='spline':
               tck_x = scipy.interpolate.bisplrep(self.grid_x, self.grid_y, dx, kx=5, ky=5)
               self.disp_x = scipy.interpolate.bisplev(self.grid_x[:,0], self.grid_y[0,:],tck_x)
               
               tck_y = scipy.interpolate.bisplrep(self.grid_x, self.grid_y, dy, kx=5, ky=5)
               self.disp_y = scipy.interpolate.bisplev(self.grid_x[:,0], self.grid_y[0,:],tck_y)
               
          else:
               self.disp_x = griddata((x, y), dx, (self.grid_x, self.grid_y), method=method)
               self.disp_y = griddata((x, y), dy, (self.grid_x, self.grid_y), method=method)



     def compute_strain_field(self):
          """Compute strain field from 
          displacement thanks to numpy"""
          #get strain fields
          dx = self.grid_x[1][0] - self.grid_x[0][0]
          dy = self.grid_y[0][1] - self.grid_y[0][0]

          
          strain_xx, strain_xy = np.gradient(self.disp_x, dx, dy, edge_order=2)
          strain_yx, strain_yy = np.gradient(self.disp_y, dx, dy, edge_order=2)

          self.strain_xx = strain_xx + .5*(np.power(strain_xx,2) + np.power(strain_yy,2))
          self.strain_yy = strain_yy + .5*(np.power(strain_xx,2) + np.power(strain_yy,2))
          self.strain_xy = .5*(strain_xy + strain_yx + strain_xx*strain_xy + strain_yx*strain_yy)
          
          
     def compute_strain_field_DA(self):
          """Compute strain field 
          from displacement field 
          thanks to a custom method 
          for large strain"""
          self.strain_xx = self.disp_x.copy(); self.strain_xx.fill(np.NAN)
          self.strain_xy = self.disp_x.copy(); self.strain_xy.fill(np.NAN)
          self.strain_yy = self.disp_x.copy(); self.strain_yy.fill(np.NAN)
          self.strain_yx = self.disp_x.copy(); self.strain_yx.fill(np.NAN)

          dx = self.grid_x[1][0] - self.grid_x[0][0]
          dy = self.grid_y[0][1] - self.grid_y[0][0]

          for i in range(self.size_x):
            for j in range(self.size_y):
                 du_dx = 0.
                 dv_dy = 0. 
                 du_dy = 0.
                 dv_dx = 0.

                 if i-1 >=0 and i+1< self.size_x:
                      du1 = (self.disp_x[i+1,j] - self.disp_x[i-1,j])/2.
                      du_dx = du1/dx
                      dv2 = (self.disp_y[i+1,j] - self.disp_y[i-1,j])/2.
                      dv_dx = dv2/dx

                 if j-1>=0 and j+1 < self.size_y:
                      dv1 = (self.disp_y[i,j+1] - self.disp_y[i,j-1])/2.
                      dv_dy = dv1/dx
                      du2 = (self.disp_x[i,j+1] - self.disp_x[i,j-1])/2.
                      du_dy = du2/dy

                 self.strain_xx[i,j] = du_dx + .5*(du_dx**2 + dv_dx**2)
                 self.strain_yy[i,j] = dv_dy + .5*(du_dy**2 + dv_dy**2)
                 self.strain_xy[i,j] = .5*(du_dy + dv_dx + du_dx*du_dy + dv_dx*dv_dy)

     def compute_strain_field_log(self):
          """Compute strain field 
          from displacement field 
          for large strain (logarithmic strain)"""
          self.strain_xx = self.disp_x.copy(); self.strain_xx.fill(np.NAN)
          self.strain_xy = self.disp_x.copy(); self.strain_xy.fill(np.NAN)
          self.strain_yy = self.disp_x.copy(); self.strain_yy.fill(np.NAN)
          self.strain_yx = self.disp_x.copy(); self.strain_yx.fill(np.NAN)


          dx = self.grid_x[1][0] - self.grid_x[0][0]
          dy = self.grid_y[0][1] - self.grid_y[0][0]
          for i in range(self.size_x):
            for j in range(self.size_y):
                 du_dx = 0.
                 dv_dy = 0. 
                 du_dy = 0.
                 dv_dx = 0.


                 if i-1 >= 0 and i+1 < self.size_x:
                      du1 = (self.disp_x[i+1,j] - self.disp_x[i-1,j])/2.
                      du_dx = du1/dx
                      dv2 = (self.disp_y[i+1,j] - self.disp_y[i-1,j])/2.
                      dv_dx = dv2/dx
                      
                 if j-1 >= 0 and j+1 < self.size_y:
                      dv1 = (self.disp_y[i,j+1] - self.disp_y[i,j-1])/2.
                      dv_dy = dv1/dx
                      du2 = (self.disp_x[i,j+1] - self.disp_x[i,j-1])/2.
                      du_dy = du2/dy
                 t11=1+2.*du_dx+du_dx**2+dv_dx**2
                 t22=1+2.*dv_dy+dv_dy**2+du_dy**2
                 t12=du_dy+dv_dx+du_dx*du_dy+dv_dx*dv_dy
                 deflog=np.log([[t11,t12],[t12,t22]])

                 self.strain_xx[i,j] = .5*deflog[0,0]
                 self.strain_yy[i,j] = .5*deflog[1,1]
                 self.strain_xy[i,j] = .5*deflog[0,1]

     def average(self, value, x_range, y_range):
          """Get the average value in 
          the specified x,y range 
          of the given field"""
          val = []
          for x in x_range:
               for y in y_range:
                    if np.isnan(value[x,y]) == False:
                         val.append(value[x,y])
          return np.average(val)

     def std(self, value, x_range, y_range):
          """Get the standard deviation value in the specified x,y range of the given field"""
          val = []
          for x in x_range:
               for y in y_range:
                    if np.isnan(value[x,y]) == False:
                         val.append(value[x,y])
          return np.std(val)




import logging
import pathlib
import tkinter as tk
import tkinter.font as tkfont
from tkinter import filedialog, messagebox, ttk

from PIL import Image, ImageTk


class TLImageProjection(tk.Toplevel):
    """top level window to project the image

    can maximize window without border (in windows)

    Args:

    """    
    fullScreen = False
    def __init__(self, parent, image_size=(800, 600)):
        super().__init__(parent)

        self._img_size = image_size 
        self.protocol('WM_DELETE_WINDOW', self.withdraw)
        # windows only (remove the minimize/maximize button)
        self.attributes('-toolwindow', True)
        self.resizable(1, 1)

        self.title('Image Projection Window (F11:maximize, Esc: Emergency Stop)')
        self.geometry(f'{image_size[0]}x{image_size[1]}')

        # layout on the root window
        self.columnconfigure(0, weight=1)
        self.columnconfigure(1, weight=1)
        self.columnconfigure(2, weight=1)

        self.__create_widgets()

    @property
    def imgsize_x(self):
        return self._img_size[0]
    @property   
    def imgsize_y(self):
        return self._img_size[1]

    def __create_widgets(self):
        # create the input frame
        # input_frame = InputFrame(self)
        # input_frame.grid(column=0, row=1)

        # set bindings
        self.bind("<F11>", func=self.toggleFullScreen)
        self.bind("<Alt-Return>", func=self.toggleFullScreen)
        self.bind("<Escape>", func=self.activateEmergencyStop)

        self.canvas = tk.Canvas(self, width=self.imgsize_y, height=self.imgsize_x,  highlightthickness=0)
        self.canvas.pack()
        self.canvas.configure(background='black')

    def update_canvas(self, pilImage):
        """main function. This requires a pilImage 

        Usage 
            pilImage  =  Image.open(IMAGE_NAME)
            self.update_canvas(pilImage=pilImage)

        Args:
            pilImage (_type_): _description_
        """        
        self.image = ImageTk.PhotoImage(pilImage)
        self.imagesprite = self.canvas.create_image(self.imgsize_y / 2, self.imgsize_x / 2, image=self.image)
        # self.imagesprite = self.canvas.create_image(0 , 0, image=self.image)
        self.canvas.image = self.imagesprite

    def toggleFullScreen(self, event):
        if self.fullScreen:
            self.deactivateFullscreen()
        else:
            self.activateFullscreen()

    def activateFullscreen(self):
        self.fullScreen = True

        # Store geometry for reset
        self._geometry_old = self.geometry()

        # Hides borders and make truly fullscreen
        self.overrideredirect(True)

        # Maximize window (Windows only). Optionally set screen geometry if you have it
        self.state("zoomed")

        # # code for linux 
        # self.geometry = self.geometry()
        # # Hides borders and make truly fullscreen
        # self.overrideredirect(True)
        # # Maximize window (Windows only). Optionally set screen geometry if you have it
        # if sys.platform.startswith('linux'):
        #     # this should work on linux
        #     self.attributes('-zoomed', True)
        # else:
        #     # this should work on macos and windows
        #     self.state("zoomed")

    def deactivateFullscreen(self):
        self.fullScreen = False
        self.state("normal")
        self.geometry(self._geometry_old)
        self.overrideredirect(False)

        # code for linux compatibility
        # if sys.platform.startswith('linux'):
        #     # this should work on linux
        #     self.root.attributes('-zoomed', False)
        # else:
        #     # this should work on macos and windows
        #     self.root.state("normal")
        
        # self.root.geometry(self.geometry)
        # self.root.overrideredirect(False)

    def activateEmergencyStop(self,event):
        """activates the emergency stop
        """        
        logging.debug('Emergency stop Activated from Image Projection window')
        self.master.model.isEmergencyStopActivated = True


import numpy as np
import matplotlib.pyplot as plt

import logging

logging.basicConfig(level=logging.DEBUG)

class PointSelector:
    """
    Select one point  and highlight it.
    Use the 'n' and 'p' keys to browse through the next and previous points
    """

    def __init__(self, fig:plt.figure, ax:plt.Axes, attach_event_callbacks:bool=True):
        """constructs the object

        Args:
            ax (ax object): it assumes that there is one line in it
        """      
        self._fig = fig   
        self._ax = ax

        #select the line and get the xy data from the 
        self._line = self._ax.lines[0]
        self.xs = self._line.get_xdata()
        self.ys = self._line.get_ydata()


        self.lastind = 0

        self.text = self._ax.text(0.05, 0.95, 'selected: none',
                            transform=self._ax.transAxes, va='top')
        self.selected, = self._ax.plot([self.xs[0]], [self.ys[0]], 'o', ms=12, alpha=0.4,
                                 color='yellow', visible=False)

        if attach_event_callbacks:
            self.attach_event_callbacks()

    @property
    def x_selected(self)->float:
        """returns x coordinate

        Returns:
            float: x coordinate
        """        
        return self.xs[self.lastind]

    @property
    def y_selected(self)->float:
        """returns y coordinate

        Returns:
            float: y coordinate
        """        
        return self.ys[self.lastind]
    
    @property
    def xy_selected(self)->tuple:
        """returns (x,y) tuple

        Returns:
            tuple: xy data
        """        
        return (self.xs[self.lastind], self.ys[self.lastind])

    @property
    def selected_index(self)->int:
        """returns selected index

        Returns:
            int: select index
        """        
        return self.lastind


    def attach_event_callbacks(self):
        """attaches the event callback on the figure object
        """        
        self._fig.canvas.mpl_connect('pick_event', self.on_pick)
        self._fig.canvas.mpl_connect('key_press_event', self.on_keypress)

    def on_keypress(self, event):
        if self.lastind is None:
            return
        if event.key not in ('n', 'p'):
            return
        if event.key == 'n':
            inc = 1
        else:
            inc = -1

        self.lastind += inc
        self.lastind = np.clip(self.lastind, 0, len(self.xs) - 1)
        self.update()

    def on_pick(self, event):

        if event.artist != self._line:
            return True

        N = len(event.ind)
        if not N:
            return True

        # the click locations
        x = event.mouseevent.xdata
        y = event.mouseevent.ydata

        distances = np.hypot(x - self.xs[event.ind], y - self.ys[event.ind])
        indmin = distances.argmin()
        dataind = event.ind[indmin]

        self.lastind = dataind
        self.update()

    def update(self):
        if self.lastind is None:
            return

        dataind = self.lastind

        self.selected.set_visible(True)
        self.selected.set_data(self.xs[dataind], self.ys[dataind])
        logging.debug(f'x:{self.xs[dataind]}, y:{ self.ys[dataind]}')
        self.text.set_text(f'selected: {dataind}' )
        self._fig.canvas.draw()


if __name__ == '__main__':
    import matplotlib.pyplot as plt
    # Fixing random state for reproducibility

    xs = np.linspace(0, 2*np.pi,100)
    ys = np.sin(xs)

    fig2, ax = plt.subplots(1, 1)
    ax.set_title('click on point to plot time series')
    line, = ax.plot(xs, ys, 'o', picker=True, pickradius=5)

    browser = PointSelector(fig=fig2, ax=(ax))

    plt.show()


#%%
import pathlib
# from dataclasses import dataclass

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd


# @dataclass
class TensileData():
    """package for loading testing data from M2 machine

    process the data and generates 
    #TODO Find machine details Imada MX2 - 2500 N

    """    
    
    def __init__(self, fname):
        self._fname = fname
        self._rawdata = pd.read_csv(filepath_or_buffer=fname, skiprows=13, names=[ 'load','displacement'])
        self._preprocess_data()

    def _preprocess_data(self):
        """preprocesses data
        
        creates calculation for each displacement point:
        - load average
        - load std
        - load min 
        - load max
        - ci for +/- 2SD
        """        
        df = self._rawdata
        df_mean = pd.concat( [
            df.groupby('displacement').mean().rename(columns={"load": 'load_avg'}), 
            df.groupby('displacement').std().rename(columns={"load": 'load_std'}),
            df.groupby('displacement').min().rename(columns={"load": 'load_min'}), 
            df.groupby('displacement').max().rename(columns={"load": 'load_max'})],axis=1)
        df_mean.eval('ci2sdm = load_avg - 2*load_std', inplace=True)
        df_mean.eval('ci2sdp = load_avg + 2*load_std', inplace=True)
        self._data = df_mean

    @property
    def load_displacement_avg_data(self):
        """returns the displacement and average data

        Returns:
            _type_: _description_
        """        
        return self._data.iloc[:,0].reset_index()

    @property
    def data_avg_table(self):
        return self._data



#%%
if __name__ == '__main__':
    FNAME = pathlib.Path('data')/"new XY 0 ABS_CNT 2%.csv"
    # %%
    td = TensileData(fname=FNAME)

    # %%

    # %%
    df_mean = td.data_avg_table
    fig, ax = plt.subplots()
    ax.plot(df_mean.index, df_mean.iloc[:,0].values, alpha=.1, label= 'avg')
    ax.plot(df_mean.index, df_mean.iloc[:,2:].values, alpha=.3, label=df_mean.iloc[:,2:].columns)# ['min', 'max', 'ci:2SD','c2'])
    # plt.ylim([155,165])
    # plt.xlim([4,8])
    ax.legend()
    #%%
    # 3, for example, is tolerance for picker i.e, how far a mouse click from
    # the plotted point can be registered to select nearby data point/points.

    def on_pick(event):
        global points
        line = event.artist
        xdata, ydata = line.get_data()
        print('selected point is:',np.array([xdata, ydata]).T)

    cid = fig.canvas.mpl_connect('pick_event', on_pick)
    # %%

    plt.show()



#%%[markdown]
# This is a version Point Selector2 (separating the selection process from the gui creation)
#%% 
import datetime
import logging
import pathlib
from dataclasses import dataclass
from tkinter import Tk
from tkinter.filedialog import askopenfilename

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from matplotlib.widgets import Button, TextBox
from npp_materialslab_tools.plottools import PointsSelector2
from npp_materialslab_tools.testing_machine import TensileData, TensileSpecimenDimensions
from npp_materialslab_tools.testing_machine.app_ElasticModGUI_v1 import  ElasticityModulusCalcs

DEVELOPMENT_FLAG = True
DEVELOPMENT_FNAME = "testingMachine/data/new XY 0 ABS_CNT 2%.csv"  

if DEVELOPMENT_FLAG:
    logging.basicConfig(level=logging.DEBUG)
else:
    logging.basicConfig(level=logging.ERROR)


class ElasticityModulusCalculatorGUIv2(object):
    """
    Like Cursor but the crosshair snaps to the nearest x, y point.
    For simplicity, this assumes that *x* is sorted.

    This is a simple state machine, with the following states:
    - 0: uninitialised (ready to select first point)
    - 1: selected first point (ready to select first point)
    - 2: selected two points (perform computation)
    
    # TODO split GUI from calculations
    """

    def __init__(self,  filename=None):
        self._res_filename = None
        self.fig, self.ax = plt.subplots(figsize=(12, 12))
        plt.subplots_adjust(bottom=0.25, top=0.98)
        self.txt = self.ax.text(0.7, 0.05, '', transform=self.ax.transAxes)
    
        
        self.load_data(filename=filename)
        self.create_plot()
        self.ps2 =  PointsSelector2(fig = self.fig, ax = self.ax)
        
        self.ps2.reset_SM()
        self._createTBs()

    def load_data(self, filename=None):
        """function that loads the data

        Args:
            filename (_type_, optional): _description_. Defaults to None.
        """ 
        if DEVELOPMENT_FLAG:
            filename = DEVELOPMENT_FNAME    
        if not filename:
            filename = "testingMachine/data/new XY 0 ABS_CNT 2%.csv"
            Tk().withdraw() # we don't want a full GUI, so keep the root window from appearing
            filename = askopenfilename(initialdir = "./")
        try:
            self._fname = pathlib.Path(filename)
            # self._tdobj = TestingData(fname=filename)

            self.emc = ElasticityModulusCalcs(tdobj=TensileData(fname=filename))
            
        except Exception as e:
            print(e)        

    @property 
    def displacement(self):
        """Delegates displacement to the ElasticityModulusCalcs

        Returns:
            _type_: _description_
        """        
        return self.emc.displacement
    @property 
    def F_Ns(self):
        """Delegates the F_Ns to the  ElasticityModulusCalcs

        Returns:
            _type_: _description_
        """        
        return self.emc.F_Ns

    def create_plot(self):
        # text location in axes coords
        self.ax.cla()
        self.ax.plot(self.displacement, self.F_Ns, 'o')
        self.ax.set_xlabel( 'Displacement [mm]')
        self.ax.set_ylabel( 'F[N]')
        self.ax.figure.canvas.draw()
        

    def _createTBs(self):
        '''
        Creates textBoxes and Reset button
        plt.axes([x, y, dx, dy])
        - x from lower left corner
        - y from lower left corner
        - dx horizontal width 
        - dy Vertical width 
        '''

        row1y,row2y =   0.12, 0.02
        col1x, col2x, col3x, col4x = 0.15, 0.35, 0.65, 0.85
        btnWidth, btnHeight = 0.10, 0.05

        self.axWidthBtn = plt.axes([col1x, row1y, btnWidth, btnHeight])
        self.tbWidth = TextBox(self.axWidthBtn, 'Width [mm]', initial='10.1')
        self.axThicknessBtn = plt.axes([col2x, row1y, btnWidth, btnHeight])
        self.tbThickness = TextBox(self.axThicknessBtn, 'Thickness [mm]', initial='3.1')
        self.axGaugeLengthBtn = plt.axes([col1x, row2y, btnWidth, btnHeight])
        self.tbGaugeLength = TextBox(self.axGaugeLengthBtn, 'G. Length [mm]', initial='90')
        
        self.tbWidth.on_submit(lambda event: self.computations())
        self.tbThickness.on_submit(lambda event: self.computations())
        
        # Reset button
        self.axResetBtn = plt.axes([col4x, row1y, btnWidth, btnHeight])
        self.btnReset = Button(self.axResetBtn, 'Reset')
        self.btnReset.on_clicked(self.on_reset)

        # Computations  button
        self.axComputationsBtn = plt.axes([col3x, row1y, btnWidth, btnHeight])
        self.btnComputations = Button(self.axComputationsBtn, 'computations')
        self.btnComputations.on_clicked(lambda event: self.computations())

        # Open new data set
        self.axOpenNewFileBtn = plt.axes([col3x, row2y, btnWidth, btnHeight])
        self.btnNewFile = Button(self.axOpenNewFileBtn, 'New file')
        self.btnNewFile.on_clicked(lambda event: self.new_file())

        # append to file button
        self.axAppendToFileBtn = plt.axes([col4x, row2y, btnWidth, btnHeight])
        self.btnAppend = Button(self.axAppendToFileBtn, 'To File')
        self.btnAppend.on_clicked(lambda event: self.appendToFile())

    def on_reset(self, event):
        self.ps2.reset_SM()
        self.create_plot()
        self.txt = self.ax.text(0.01, 0.95, self._fname.stem, transform=self.ax.transAxes)

    def new_file(self):
        """loads new file
        """        
        self.load_data(filename=None)
        self.reset_SM()

    def appendToFile(self):
        if self._sm != 2:
            print ('Cannot estimate E yet!')
            return 
        if not self._res_filename:
            self._res_filename = datetime.datetime.now().strftime("res%Y%m%d%H%M%S.txt")
            #self._res_filename = 'results_log.txt'
            with open(self._res_filename, 'w') as file:
                file.write('{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}\n'.format('Timestamp','Filename','Width', 'Thickness', 'start_Ind', 'end_Ind', 'E_GPA_pt', 'E_GPA_lrg'))
       
        with open(self._res_filename, 'a') as file:
            E_GPa_pt = self._compute_mod_elasticity()
            E_GPa_lnr = self._compute_mod_elasticity_lnr(self._points_collected[1].indx, self._points_collected[2].indx)
            file.write('{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}\n'.format(datetime.datetime.now().strftime("%Y%m%d-%H%M%S"), self._fname.stem ,self._dim_w, self._dim_t, self._points_collected[1].indx, self._points_collected[2].indx, E_GPa_pt, E_GPa_lnr))
       
    

    def get_specimenDimensions(self):
        ''' calcate crossection'''
        self._dim_w = float(self.tbWidth.text)
        self._dim_t = float(self.tbThickness.text)
        self._dim_l = float(self.tbGaugeLength.text)

        self._specimenDimensions = TensileSpecimenDimensions(gauge_length_mm=self._dim_l, thickness_mm= self._dim_t, width_mm=self._dim_w)
        return self._specimenDimensions

    def computations(self):
        specimenDimensions= self.get_specimenDimensions()
        self.exxs = self.displacement / specimenDimensions.gauge_length_mm
        res= self.emc.computations(specimen=specimenDimensions, points_collected = self.ps2._points_collected)
        E_GPa_pt = res['E_MPa_simple']
        E_GPa_lnr = res['E_MPa_lsq']
        #output to plot and console
        start_ind = self.ps2._points_collected[1].indx
        end_ind = self.ps2._points_collected[2].indx
        self._l_selected = self.ax.plot(self.displacement[start_ind:end_ind], self.F_Ns[start_ind:end_ind], 'r')
        self.txt = self.ax.text(0.7, 0.05, '', transform=self.ax.transAxes)
        self.txt.set_text('E = {:.2f}[Mpa]'.format (E_GPa_lnr))
        # console
        print('E (pt): {:0.2f}[MPa]       E (linear regression): {:0.2f}[MPa] '.format(E_GPa_pt, E_GPa_lnr))



#%%
if __name__ == "__main__":

    # snap_cursor = ElasticityModulusCalculator('../Results/20190327-tests/D00_02.xlsx')
    snap_cursor = ElasticityModulusCalculatorGUIv2()

    plt.show()


# from .appElasticityCalculator import ElasticityModulusCalcs
from npp_materialslab_tools.testing_machine.tensile_data import TensileData
from npp_materialslab_tools.testing_machine.specimen_classes import TensileSpecimenDimensions


import pathlib

import numpy as np
import pandas as pd
from matplotlib import pyplot as plt
import logging


def convert_labview_to_metainfo(in_fname:pathlib.Path,out_fname:pathlib.Path, img_ext:str="png"):
    """coverts the output file from labview to the "meta-data.txt" required by pydic

    Args:
        in_fname (pathlib.Path): input filename
        out_fname (pathlib.Path): output filename 
        img_ext (str): Image file extension 
    """    
    dfcols = ['ImgNo', "Timestamp", 'ElapsedTime']
    df = pd.read_csv(in_fname, delimiter="\t", header=None, names=dfcols)

    df['filename'] = df.ImgNo.apply(lambda x: f"Cam_{x:05d}.{img_ext}")

    df = df.loc[:, ['filename', "ElapsedTime"]]
    # The following mocks up the force 
    df['force'] = df['ElapsedTime'].round(decimals=0)

    df.columns= ['file', 'time(s)', 'force(N)']

    df.to_csv(out_fname, index=False, sep='\t')


def plot_synced_graph(time_offset, dic_df:pd.DataFrame, ut_df:pd.DataFrame):
    """plots the dic and tensile data to see if they correlate well
    
    Two plots are made:
    - one plot with the normalised force and the normalised strain
    - plot:
        - the force diff normalised wrt abs(max(force diff)
        - the strain diff normalised wrt abs(max(strain diff))

    TODO: I could use cross-correlation but this would require similar timestep. 
        
    Args:
        time_offset (float): Time offset in s
        dic_df (pd.DataFrame): dataframe obtained from dic
        ut_df (pd.DataFrame): dataframe obtained from imada
    """    
    # time_offset = 1.55
    dic_df["time_synced"] = dic_df["time(s)"]-time_offset
    # axs = [axs]
    ts_ut = ut_df.time_s
    Fs_ut = ut_df.force_N
    ts_dic = dic_df["time_synced"]
    exx_dic = dic_df.e_xx
    fig, axs = plt.subplots(ncols=1,nrows=2,sharex=True)

    # plot 1
    axs[0].plot(ts_ut, Fs_ut/Fs_ut.max(), '.', label ="Normalised Force")
    axs[0].plot(ts_dic[:-1], exx_dic[:-1]/exx_dic[:-1].max(), '.',label ="Normalised strain ")
    axs[0].set_title(f"Normalised Forces (from Imada) and Strains (from dic)\n Used to determine time offset: {time_offset} (s)")

    # plot 2 with normalised diffs (the  )
    axs[1].plot(ts_ut[:-1],np.abs(np.diff(Fs_ut))/np.abs(np.diff(Fs_ut)).max(), label ="force diff")
    axs[1].plot(ts_dic[:-1], np.abs(np.diff(exx_dic))/np.abs(np.diff(exx_dic)).max(),label ="Normalised strain ")
    axs[1].set_xlabel("Time (s)")



# %% [markdown]
# # Resampling
# ## obsolete attempt with pandas.resample
# The following is an failed attempt to resampling the dataframe 
# on sub second time scale. 
# 
# Issues:
# - I could not obtain meaningful values when I resampled
#%%
# index_td = pd.to_timedelta(df_dic['time_synced'], unit ="s")
# # index_td = pd.to_datetime(df_dic['time_synced'], unit ="s")
# df_dict = df_dic.set_index(index_td )
# df_dict.resample('100ms',origin=0).interpolate('linear')


def resample_dic_df(df_dic:pd.DataFrame,  ts:np.ndarray)->pd.DataFrame:
    """resamples the DIC analysis dataframe, on another time vector.

    Args:
        df_dic (pd.DataFrame): DIC Dataframe 
        ts (np.ndarray): new time vector. eg: np.arange(0,14,step=0.5).

    Returns:
        pd.DataFrame: new DataFrame with results 
    """
    dss = []
    for columnname in df_dic.iloc[:,:4]:

        interp_data = np.interp(ts, df_dic['time_synced'],df_dic[columnname])
        ds_new = pd.Series(interp_data, index = ts, name=columnname)
        dss.append(ds_new)
    res_df = pd.concat(dss,axis=1)
    return res_df 

def resample_ut(df_ut:pd.DataFrame,  ts:np.ndarray)->pd.DataFrame:
    """resamples the IMADA M testing machine results, on another time vector.

    Args:
        df_ut (pd.DataFrame): imada ut machine Dataframe (columms: [force_N, disp_mm, time_s])
        ts (np.ndarray): new time vector. eg: np.arange(0,14,step=0.5).

    Returns:
        pd.DataFrame: new DataFrame with results 
    """
    assert all([df_ut.columns[k]== v  for  k, v in enumerate(['force_N', 'disp_mm', 'time_s'])]), "Columns in input Df should be ['force_N', 'disp_mm', 'time_s']"
    dss = []
    for columnname in df_ut.iloc[:,:2]:
        logging.debug(columnname)
        interp_data = np.interp(ts, df_ut['time_s'],df_ut[columnname])
        ds_new = pd.Series(interp_data, index = ts, name=columnname)
        dss.append(ds_new)
    res_df = pd.concat(dss,axis=1)
    return res_df




import logging
import pathlib
import tkinter as tk
import tkinter.font as tkfont
from tkinter import filedialog, messagebox, ttk
import numpy as np


import matplotlib.pyplot as plt
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg


class TLMatplotlibGraph(tk.Toplevel):
    """top level window to display a graph

    can maximize window without border (in windows)

    Args:

    """    
    fullScreen = False
    def __init__(self, parent, **kwargs):
        super().__init__(parent)

        self.protocol('WM_DELETE_WINDOW', self.withdraw)
        # windows only (remove the minimize/maximize button)
        self.attributes('-toolwindow', True)
        self.resizable(1, 1)

        self.title('Plot Window (F11:maximize, Esc: Emergency Stop)')
        self.geometry('500x300')

        # layout on the root window
        self.columnconfigure(0, weight=1)
        self.columnconfigure(1, weight=1)
        self.columnconfigure(2, weight=1)

        self.__create_widgets()

    def __create_widgets(self):
        # create the input frame
        # input_frame = InputFrame(self)
        # input_frame.grid(column=0, row=1)

        # set bindings
        self.bind("<F11>", func=self.toggleFullScreen)
        self.bind("<Alt-Return>", func=self.toggleFullScreen)
        self.bind("<Escape>", func=self.activateEmergencyStop)


        self.fig=plt.figure(figsize=(8,8))
        self._ax=self.fig.add_axes([0.1,0.1,0.8,0.8],polar=False)
        self.canvas = FigureCanvasTkAgg(self.fig,master=self)
        self.canvas.get_tk_widget().grid(row=0,column=1)
        self.canvas.draw()


    @property
    def ax(self):
        return self._ax


    def toggleFullScreen(self, event):
        if self.fullScreen:
            self.deactivateFullscreen()
        else:
            self.activateFullscreen()

    def activateFullscreen(self):
        self.fullScreen = True

        # Store geometry for reset
        self._geometry_old = self.geometry()

        # Hides borders and make truly fullscreen
        self.overrideredirect(True)

        # Maximize window (Windows only). Optionally set screen geometry if you have it
        self.state("zoomed")

        # # code for linux 
        # self.geometry = self.geometry()
        # # Hides borders and make truly fullscreen
        # self.overrideredirect(True)
        # # Maximize window (Windows only). Optionally set screen geometry if you have it
        # if sys.platform.startswith('linux'):
        #     # this should work on linux
        #     self.attributes('-zoomed', True)
        # else:
        #     # this should work on macos and windows
        #     self.state("zoomed")

    def deactivateFullscreen(self):
        self.fullScreen = False
        self.state("normal")
        self.geometry(self._geometry_old)
        self.overrideredirect(False)

        # code for linux compatibility
        # if sys.platform.startswith('linux'):
        #     # this should work on linux
        #     self.root.attributes('-zoomed', False)
        # else:
        #     # this should work on macos and windows
        #     self.root.state("normal")
        
        # self.root.geometry(self.geometry)
        # self.root.overrideredirect(False)

    def activateEmergencyStop(self,event):
        """activates the emergency stop
        """        
        logging.debug('Emergency stop Activated from Image Projection window')
        self.master.destroy()

if __name__ == "__main__":
    root = tk.Tk()
    root.withdraw()
    tlmg = TLMatplotlibGraph(parent=root)
    x = np.linspace(0, np.pi*2 ,100)
    y = np.sin(x)
    tlmg.ax.plot(x,y)
    tlmg.mainloop()



import pytest
from .app_ElasticModGUI_v1 import TensileSpecimenDimensions

def test_one():
    pass

def test_SpecimenDimensions():
    sd = TensileSpecimenDimensions(width_mm=10, thickness_mm=3, gauge_length_mm=100)
    assert sd.width_mm ==10
    assert sd.thickness_mm==3
    assert sd.gauge_length_mm==100
    assert sd.csArea_mm2==30


import pathlib
import pandas as pd

def read_imada_csv(fname:pathlib.Path|str, decimation:int=100)->pd.DataFrame:
    """Reads imada csv file and returns a decimated Dataframe

    with columns 	
        - force_N	
        - disp_mm	
        - time_s

    Args:
        fname (pathlib.Path|str): imada csv export (from z3a file)
        decimation (int, optional): reduce the data by the decimation factor. Defaults to 100.

    Returns:
        pd.DataFrame: a Dataframe containing columns = ["force_N", "disp_mm", "time_s"]
    """
    COLNAMES = ["force_N", "disp_mm", "time_s"]
    # obtain the meta data 
    df_ut_meta = pd.read_csv(fname, sep="=",nrows=13, header=None, index_col=0, names=["Description", "Value"])
    RECORDING_DT = float(df_ut_meta.loc['RECORDING RATE ','Value'].strip()[:-1])
    # obtain the dataframe of the test
    df_ut = pd.read_csv(fname, skiprows=13, names = ["Force (N)", "Disp (mm)"])
    df_ut['Time (s)'] = df_ut.index*RECORDING_DT
    # df_ut.head()
    df_decimated = df_ut.iloc[::decimation,:]
    df_decimated.columns = COLNAMES
    return df_decimated


#%% 
import logging
from dataclasses import dataclass
from tkinter import Tk
from tkinter.filedialog import askopenfilename

import matplotlib.pyplot as plt
import numpy as np
from npp_materialslab_tools import TensileData
from npp_materialslab_tools.plottools import Cursor

# DEVELOPMENT_FLAG = True
# DEVELOPMENT_FNAME = "testingMachine/data/new XY 0 ABS_CNT 2%.csv"  

# if DEVELOPMENT_FLAG:
#     logging.basicConfig(level=logging.DEBUG)
# else:
#     logging.basicConfig(level=logging.ERROR)
class PointsSelector2():
    def __init__(self, fig, ax):

        self.fig = fig
        self.ax = ax
        self.fig.canvas.mpl_connect('button_press_event', self.mouse_click)

        self.reset_SM()
        
    def reset_SM(self):
        ''' resets state machine status'''
        self._points_collected = {}
        self._sm = 0
        # self.txt.set_text('')

        # self.ax.cla()

    def _calc_x_y_ind(self, x_event, y_event ):
        ''' calculates the x, y and index from the event data

        It does that by finding the closest point.

        Callers: mouse_click
        '''
        try:
            line = self.ax.lines[0]
            self.xdata = line.get_xdata()
            self._F_Ns = line.get_ydata()
            indx = min(np.searchsorted(self.xdata, x_event), len(self.xdata) - 1)
            x = self.xdata[indx]
            y = self._F_Ns[indx]
            return x, y, indx   
        except:
            return 

    def mouse_click(self, event):
        ''' change the state machine on each click
        ''' 
        if not event.inaxes:
            ''' only continue when a point is picked'''
            return

        # update state machine
        if self._sm == 0:
            crsrID  = self._sm+1
            x,y, indx = self._calc_x_y_ind(event.xdata, event.ydata)
            # update the line positions
            self._points_collected[crsrID] = Cursor(self.ax, crsrID, x, y, indx)
            self._points_collected[crsrID].plot_cursor()
            self._sm = 1
        elif self._sm == 1:
            crsrID  = self._sm+1
            x,y, indx = self._calc_x_y_ind(event.xdata, event.ydata)
            # update the line positions
            self._points_collected[crsrID] = Cursor(self.ax, crsrID, x, y, indx)
            self._points_collected[crsrID].plot_cursor()
            
            self._sm = 2            
        if self._sm ==2:
            logging.debug ("ready to plot")
            # self.computations()
            pass
        else:
            print ("State status: {} | x={:1.2f}, y={:1.2f}".format(self._sm,x,y))
        self.ax.figure.canvas.draw()



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

requirements = [
    #'opencv',
    'scipy',
    'numpy',
    'matplotlib',
    'seaborn',
    'pandas',
    'openpyxl',
    'ipykernel',
    'jupyter'
 ]

test_requirements = [
    'pytest',
    # 'pytest-pep8',
    # 'pytest-cov',
]


setuptools.setup(
    name="npp_materialslab_tools", # Replace with your own username
    version="0.0.1",
    author="N. Papadakis",
    author_email="npapnet@gmail.com",
    description="A package for the material lab tools",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=requirements,
    tests_require=test_requirements,
    python_requires='>=3.8',
)


class Cursor():
    def __init__(self, ax, ID, x, y, indx):
        '''
        '''
        self.ID = ID
        self.ax = ax # make cursor aware of the ax that's been plotted to. 
        self.x = x
        self.y = y
        self.indx = indx
        # defaults
        self.col = 'k'
        self._linestyle = 'dashed'

    def __str__(self):
        return "ID:{}, (x:{}, y:{}) @ ind:{}".format(self.ID, self.x, self.y, self.indx)

    def plot_cursor(self, new = True):
        ''' Plots a cursor 
        '''
        if new:
            self.lx = self.ax.axhline(self.y, color=self.col, linestyle=self._linestyle)  # the horiz line
            self.ly = self.ax.axvline(self.x, color=self.col, linestyle=self._linestyle)  # the vert line
        else:
            self.lx.set_ydata(self.y)
            self.ly.set_xdata(self.x)

    def remove_line(self):
            self.ax.lines.remove(self.lx)
            self.ax.lines.remove(self.ly)        
            self.lx = None
            self.ly = None


